import numpy as np
from abc import abstractmethod

from .farm_opt_problem import FarmOptProblem
from foxes.models.turbine_models import SetFarmVars
from foxes.config import config


class FarmVarsProblem(FarmOptProblem):
    """
    Abstract base class for models that optimize
    farm variables.

    :group: opt.core

    """

    def initialize(self, pre_rotor_vars, post_rotor_vars, verbosity=1, **kwargs):
        """
        Initialize the object.

        Parameters
        ----------
        pre_rotor_vars: list of str or dict
            The pre_rotor farm variables. If dict, then
            key: sub-model str, value: var names as list of str
        post_rotor_vars: list of str or dict
            The post_rotor farm variables. If dict, then
            key: sub-model str, value: var names as list of str
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for super class init

        """
        self._vars_pre = {}
        self._vars_post = {}
        if isinstance(pre_rotor_vars, dict):
            self._vars_pre = {m: v for m, v in pre_rotor_vars.items() if len(v)}
        elif len(pre_rotor_vars):
            self._vars_pre = {self.name: pre_rotor_vars}
        if isinstance(post_rotor_vars, dict):
            self._vars_post = {m: v for m, v in post_rotor_vars.items() if len(v)}
        elif len(post_rotor_vars):
            self._vars_post = {self.name: post_rotor_vars}

        cnt = 0
        for src, pre in zip((self._vars_pre, self._vars_post), (True, False)):
            for mname, vrs in src.items():
                if mname in self.algo.mbook.turbine_models:
                    m = self.algo.mbook.turbine_models[mname]
                    if not isinstance(m, SetFarmVars):
                        raise KeyError(
                            f"FarmOptProblem '{self.name}': Turbine model entry '{mname}' already exists in model book, and is not of type SetFarmVars"
                        )
                    elif m.pre_rotor != pre:
                        raise ValueError(
                            f"FarmOptProblem '{self.name}': Turbine model entry '{mname}' exists in model book, and disagrees on pre_rotor = {pre}"
                        )
                else:
                    self.algo.mbook.turbine_models[mname] = SetFarmVars(pre_rotor=pre)

                found = False
                for t in self.algo.farm.turbines:
                    if mname in t.models:
                        found = True
                        break
                if not found:
                    raise ValueError(
                        f"FarmOptProblem '{self.name}': Missing entry '{mname}' among any of the turbine models"
                    )
                cnt += len(vrs)
        if not cnt:
            raise ValueError(
                f"Problem '{self.name}': Neither pre_rotor_vars not post_rotor_vars containing variables"
            )

        super().initialize(verbosity=verbosity, **kwargs)

    @abstractmethod
    def opt2farm_vars_individual(self, vars_int, vars_float):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_vars_int,)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_vars_float,)

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_states, n_sel_turbines)

        """
        pass

    @abstractmethod
    def opt2farm_vars_population(self, vars_int, vars_float, n_states):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_pop, n_vars_int)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_pop, n_vars_float)
        n_states: int
            The number of original (non-pop) states

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_pop, n_states, n_sel_turbines)

        """
        pass

    def update_problem_individual(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        """
        super().update_problem_individual(vars_int, vars_float)

        # prepare:
        n_states = self._org_n_states
        fvars = self.opt2farm_vars_individual(vars_int, vars_float)

        # update turbine model that sets vars to opt values:
        for src in (self._vars_pre, self._vars_post):
            for mname, vrs in src.items():
                model = self.algo.mbook.turbine_models[mname]
                model.reset()
                for v in vrs:
                    vals = fvars.pop(v)
                    if self.all_turbines:
                        model.add_var(v, vals)
                    else:
                        data = np.zeros(
                            (n_states, self.algo.n_turbines), dtype=config.dtype_double
                        )
                        data[:, self.sel_turbines] = vals
                        model.add_var(v, data)

        if len(fvars):
            raise KeyError(
                f"Problem '{self.name}': Too many farm vars from opt2farm_vars_individual: {list(fvars.keys())}"
            )

    def update_problem_population(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float,)

        """
        super().update_problem_population(vars_int, vars_float)

        # prepare:
        n_pop = len(vars_float)
        n_states = self._org_n_states
        n_pstates = n_states * n_pop
        fvars = self.opt2farm_vars_population(vars_int, vars_float, n_states)

        # update turbine model that sets vars to opt values:
        for src in (self._vars_pre, self._vars_post):
            for mname, vrs in src.items():
                model = self.algo.mbook.turbine_models[mname]
                model.reset()
                for v in vrs:
                    vals = fvars.pop(v)
                    shp0 = list(vals.shape)
                    shp1 = [n_pstates] + shp0[2:]
                    if self.all_turbines:
                        model.add_var(v, vals.reshape(shp1))
                    else:
                        data = np.zeros(
                            (n_pstates, self.algo.n_turbines), dtype=config.dtype_double
                        )
                        data[:, self.sel_turbines] = vals.reshape(shp1)
                        model.add_var(v, data)
                        del data

        if len(fvars):
            raise KeyError(
                f"Problem '{self.name}': Too many farm vars from opt2farm_vars_population: {list(fvars.keys())}"
            )



import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist
from iwopy import Problem

from foxes.config import config


class GeomRegGrids(Problem):
    """
    A regular grid within a boundary geometry.

    This optimization problem does not involve
    wind farms.

    Attributes
    ----------
    boundary: foxes.utils.geom2d.AreaGeometry
        The boundary geometry
    min_dist: float
        The minimal distance between points
    n_grids: int
        The number of grids
    n_max: int
        The maximal number of points
    n_row_max: int
        The maximal number of points in a row
    max_dist: float
        The maximal distance between points
    D: float
        The diameter of circle fully within boundary

    :group: opt.problems.layout.geom_layouts

    """

    def __init__(
        self,
        boundary,
        min_dist,
        n_grids,
        n_max=None,
        n_row_max=None,
        max_dist=None,
        D=None,
    ):
        """
        Constructor.

        Parameters
        ----------
        boundary: foxes.utils.geom2d.AreaGeometry
            The boundary geometry
        min_dist: float
            The minimal distance between points
        n_grids: int
            The number of grids
        n_max: int, optional
            The maximal number of points
        n_row_max: int, optional
            The maximal number of points in a row
        max_dist: float, optional
            The maximal distance between points
        D: float, optional
            The diameter of circle fully within boundary

        """
        super().__init__(name="geom_reg_grids")

        self.boundary = boundary
        self.n_grids = n_grids
        self.n_max = n_max
        self.n_row_max = n_row_max
        self.min_dist = float(min_dist)
        self.max_dist = float(max_dist) if max_dist is not None else max_dist
        self.D = D

        self._NX = [f"nx{i}" for i in range(self.n_grids)]
        self._NY = [f"ny{i}" for i in range(self.n_grids)]
        self._OX = [f"ox{i}" for i in range(self.n_grids)]
        self._OY = [f"oy{i}" for i in range(self.n_grids)]
        self._DX = [f"dx{i}" for i in range(self.n_grids)]
        self._DY = [f"dy{i}" for i in range(self.n_grids)]
        self._ALPHA = [f"alpha{i}" for i in range(self.n_grids)]

    def initialize(self, verbosity=1):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(verbosity)

        pmin = self.boundary.p_min()
        pmax = self.boundary.p_max()
        self._span = pmax - pmin
        self._diag = np.linalg.norm(self._span)
        self.max_dist = self._diag if self.max_dist is None else self.max_dist
        self._nrow = self.n_row_max
        if self.n_row_max is None:
            if self.n_max is None:
                self._nrow = int(self._diag / self.min_dist)
                if self._nrow * self.min_dist < self._diag:
                    self._nrow += 1
                self._nrow += 1
            else:
                self._nrow = self.n_max
        if self.n_max is None:
            self.n_max = self.n_grids * self._nrow**2
        elif self.n_max <= self._nrow:
            self._nrow = self.n_max
        self._pmin = pmin - self._diag - self.min_dist
        self._pmax = pmax + self.min_dist

        if verbosity > 0:
            print(f"Grid data:")
            print(f"  pmin        = {self._pmin}")
            print(f"  pmax        = {self._pmax}")
            print(f"  min dist    = {self.min_dist}")
            print(f"  max dist    = {self.max_dist}")
            print(f"  n row max   = {self._nrow}")
            print(f"  n max       = {self.n_max}")
            print(f"  n grids     = {self.n_grids}")

        self.apply_individual(self.initial_values_int(), self.initial_values_float())

    def var_names_int(self):
        """
        The names of int variables.

        Returns
        -------
        names: list of str
            The names of the int variables

        """
        return list(np.array([self._NX, self._NY]).T.flat)

    def initial_values_int(self):
        """
        The initial values of the int variables.

        Returns
        -------
        values: numpy.ndarray
            Initial int values, shape: (n_vars_int,)

        """
        return np.full(self.n_grids * 2, 2, dtype=config.dtype_int)

    def min_values_int(self):
        """
        The minimal values of the integer variables.

        Use -self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal int values, shape: (n_vars_int,)

        """
        return np.ones(self.n_grids * 2, dtype=config.dtype_int)

    def max_values_int(self):
        """
        The maximal values of the integer variables.

        Use self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal int values, shape: (n_vars_int,)

        """
        return np.full(self.n_grids * 2, self._nrow, dtype=config.dtype_int)

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        return list(
            np.array([self._OX, self._OY, self._DX, self._DY, self._ALPHA]).T.flat
        )

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        n = 5
        vals = np.zeros((self.n_grids, n), dtype=config.dtype_double)
        vals[:, :2] = self._pmin + self._diag + self.min_dist + self._span / 2
        vals[:, 2:4] = 2 * self.min_dist
        vals[:, 5:] = 0
        return vals.reshape(self.n_grids * n)

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        n = 5
        vals = np.zeros((self.n_grids, n), dtype=config.dtype_double)
        vals[:, :2] = self._pmin
        vals[:, 2:4] = self.min_dist
        vals[:, 5:] = -self._diag / 3
        return vals.reshape(self.n_grids * n)

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        n = 5
        vals = np.zeros((self.n_grids, n), dtype=config.dtype_double)
        vals[:, :2] = self._pmax
        vals[:, 2:4] = self.max_dist
        vals[:, 4] = 90.0
        vals[:, 5:] = self._diag / 3
        return vals.reshape(self.n_grids * n)

    def apply_individual(self, vars_int, vars_float):
        """
        Apply new variables to the problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        vint = vars_int.reshape(self.n_grids, 2)
        vflt = vars_float.reshape(self.n_grids, 5)
        nx = vint[:, 0]
        ny = vint[:, 1]
        ox = vflt[:, 0]
        oy = vflt[:, 1]
        dx = vflt[:, 2]
        dy = vflt[:, 3]
        a = np.deg2rad(vflt[:, 4])
        s = vflt[:, 5:]
        n_points = self.n_max

        nax = np.stack([np.cos(a), np.sin(a), np.zeros_like(a)], axis=-1)
        naz = np.zeros_like(nax)
        naz[:, 2] = 1
        nay = np.cross(naz, nax)

        valid = np.zeros(n_points, dtype=bool)
        pts = np.full((n_points, 2), np.nan, dtype=config.dtype_double)
        n0 = 0
        for gi in range(self.n_grids):
            n = nx[gi] * ny[gi]
            n1 = n0 + n

            if n1 <= n_points:
                qts = pts[n0:n1].reshape(nx[gi], ny[gi], 2)
            else:
                qts = np.zeros((nx[gi], ny[gi], 2), dtype=config.dtype_double)

            qts[:, :, 0] = ox[gi]
            qts[:, :, 1] = oy[gi]
            qts[:] += (
                np.arange(nx[gi])[:, None, None] * dx[gi] * nax[gi, None, None, :2]
            )
            qts[:] += (
                np.arange(ny[gi])[None, :, None] * dy[gi] * nay[gi, None, None, :2]
            )
            qts = qts.reshape(n, 2)

            if n1 > n_points:
                n1 = n_points
                qts = qts[: (n1 - n0)]
                pts[n0:] = qts

            # set out of boundary points invalid:
            if self.D is None:
                valid[n0:n1] = self.boundary.points_inside(qts)
            else:
                valid[n0:n1] = self.boundary.points_inside(qts) & (
                    self.boundary.points_distance(qts) >= self.D / 2
                )

            # set points invalid which are too close to other grids:
            if n0 > 0:
                dists = cdist(qts, pts[:n0])
                valid[n0:n1][np.any(dists < self.min_dist, axis=1)] = False

            n0 = n1
            if n0 >= n_points:
                break

        return pts, valid

    def apply_population(self, vars_int, vars_float):
        """
        Apply new variables to the problem,
        for a whole population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        n_pop = vars_int.shape[0]
        vint = vars_int.reshape(n_pop, self.n_grids, 2)
        vflt = vars_float.reshape(n_pop, self.n_grids, 5)
        nx = vint[:, :, 0]
        ny = vint[:, :, 1]
        ox = vflt[:, :, 0]
        oy = vflt[:, :, 1]
        dx = vflt[:, :, 2]
        dy = vflt[:, :, 3]
        a = np.deg2rad(vflt[:, :, 4])
        s = vflt[:, :, 5:]
        n_points = self.n_max

        nax = np.stack([np.cos(a), np.sin(a), np.zeros_like(a)], axis=-1)
        naz = np.zeros_like(nax)
        naz[:, :, 2] = 1
        nay = np.cross(naz, nax)

        valid = np.zeros((n_pop, n_points), dtype=bool)
        pts = np.full((n_pop, n_points, 2), np.nan, dtype=config.dtype_double)
        for pi in range(n_pop):
            n0 = 0
            for gi in range(self.n_grids):
                n = nx[pi, gi] * ny[pi, gi]
                n1 = n0 + n

                if n1 <= n_points:
                    qts = pts[pi, n0:n1].reshape(nx[pi, gi], ny[pi, gi], 2)
                else:
                    qts = np.zeros(
                        (nx[pi, gi], ny[pi, gi], 2), dtype=config.dtype_double
                    )

                qts[:, :, 0] = ox[pi, gi]
                qts[:, :, 1] = oy[pi, gi]
                qts[:] += (
                    np.arange(nx[pi, gi])[:, None, None]
                    * dx[pi, gi]
                    * nax[pi, gi, None, None, :2]
                )
                qts[:] += (
                    np.arange(ny[pi, gi])[None, :, None]
                    * dy[pi, gi]
                    * nay[pi, gi, None, None, :2]
                )
                qts = qts.reshape(n, 2)

                if n1 > n_points:
                    n1 = n_points
                    qts = qts[: (n1 - n0)]
                    pts[pi, n0:] = qts

                # set out of boundary points invalid:
                if self.D is None:
                    valid[pi, n0:n1] = self.boundary.points_inside(qts)
                else:
                    valid[pi, n0:n1] = self.boundary.points_inside(qts) & (
                        self.boundary.points_distance(qts) >= self.D / 2
                    )

                # set points invalid which are too close to other grids:
                if n0 > 0:
                    dists = cdist(qts, pts[pi, :n0])
                    valid[pi, n0:n1][np.any(dists < self.min_dist, axis=1)] = False

                n0 = n1

                if n0 >= n_points:
                    break

        return pts, valid

    def get_fig(
        self, xy=None, valid=None, ax=None, title=None, true_circle=True, **bargs
    ):
        """
        Return plotly figure axis.

        Parameters
        ----------
        xy: numpy.ndarary, optional
            The xy coordinate array, shape: (n_points, 2)
        valid: numpy.ndarray, optional
            Boolean array of validity, shape: (n_points,)
        ax: pyplot.Axis, optional
            The figure axis
        title: str, optional
            The figure title
        true_circle: bool
            Draw points as circles with diameter self.D
        bars: dict, optional
            The boundary plot arguments

        Returns
        -------
        ax: pyplot.Axis
            The figure axis

        """
        if ax is None:
            __, ax = plt.subplots()

        hbargs = {"fill_mode": "inside_lightgray"}
        hbargs.update(bargs)
        self.boundary.add_to_figure(ax, **hbargs)

        if xy is not None:
            if valid is not None:
                xy = xy[valid]
            if not true_circle or self.D is None:
                ax.scatter(xy[:, 0], xy[:, 1], color="orange")
            else:
                for x, y in xy:
                    ax.add_patch(
                        plt.Circle((x, y), self.D / 2, color="blue", fill=True)
                    )

        ax.set_aspect("equal", adjustable="box")
        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")

        if title is None:
            if xy is None:
                title = f"Optimization area"
            else:
                l = len(xy) if xy is not None else 0
                dists = cdist(xy, xy)
                np.fill_diagonal(dists, 1e20)
                title = f"N = {l}, min_dist = {np.min(dists):.1f} m"
        ax.set_title(title)

        return ax



import numpy as np
from iwopy import Problem

from foxes.config import config

from .pop_states import PopStates


class FarmOptProblem(Problem):
    """
    Abstract base class of wind farm optimization problems.

    Attributes
    ----------
    algo: foxes.core.Algorithm
        The algorithm
    calc_farm_args: dict
        Additional parameters for algo.calc_farm()
    points : numpy.ndarray
        The probe points, shape: (n_states, n_points, 3)

    :group: opt.core

    """

    def __init__(
        self,
        name,
        algo,
        sel_turbines=None,
        calc_farm_args={},
        points=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        name: str
            The problem's name
        algo: foxes.core.Algorithm
            The algorithm
        sel_turbines: list of int, optional
            The turbines selected for optimization,
            or None for all
        calc_farm_args: dict
            Additional parameters for algo.calc_farm()
        points : numpy.ndarray, optional
            The probe points, shape: (n_states, n_points, 3)
        kwargs: dict, optional
            Additional parameters for `iwopy.Problem`

        """
        super().__init__(name, **kwargs)

        self.algo = algo
        self.calc_farm_args = calc_farm_args
        self.points = points

        self._sel_turbines = sel_turbines
        self._count = None

    @property
    def farm(self):
        """
        The wind farm

        Returns
        -------
        foxes.core.WindFarm :
            The wind farm

        """
        return self.algo.farm

    @property
    def sel_turbines(self):
        """
        The selected turbines

        Returns
        -------
        list of int :
            Indices of the selected turbines

        """
        return (
            self._sel_turbines
            if self._sel_turbines is not None
            else list(range(self.farm.n_turbines))
        )

    @property
    def n_sel_turbines(self):
        """
        The numer of selected turbines

        Returns
        -------
        int :
            The numer of selected turbines

        """
        return len(self.sel_turbines)

    @property
    def all_turbines(self):
        """
        Flag for all turbines optimization

        Returns
        -------
        bool :
            True if all turbines are subject to optimization

        """
        return len(self.sel_turbines) == self.algo.n_turbines

    @property
    def counter(self):
        """
        The current value of the application counter

        Returns
        -------
        int :
            The current value of the application counter

        """
        return self._count

    @classmethod
    def tvar(cls, var, turbine_i):
        """
        Gets turbine variable name

        Parameters
        ----------
        var: str
            The variable name
        turbine_i: int
            The turbine index

        Returns
        -------
        str :
            The turbine variable name

        """
        return f"{var}_{turbine_i:04d}"

    @classmethod
    def parse_tvar(cls, tvr):
        """
        Parse foxes variable name and turbine index
        from turbine variable

        Parameters
        ----------
        tvr: str
            The turbine variable name

        Returns
        -------
        var: str
            The foxes variable name
        turbine_i: int
            The turbine index

        """
        t = tvr.split("_")
        return t[0], int(t[1])

    def initialize(self, verbosity=1):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        if not self.algo.initialized:
            self.algo.initialize()
        self._org_states_name = self.algo.states.name
        self._org_n_states = self.algo.n_states
        self._org_weights = self.algo.states.weights(self.algo)

        self.algo.finalize()

        self._count = 0

        super().initialize(verbosity)

    def _reset_states(self, states):
        """
        Reset the states in the algorithm
        """
        if states is not self.algo.states:
            if self.algo.initialized:
                self.algo.finalize()
            self.algo.states = states

    def update_problem_individual(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        """
        # reset states, if needed:
        if isinstance(self.algo.states, PopStates):
            self._reset_states(self.algo.states.states)
            self.algo.n_states = self._org_n_states

    def update_problem_population(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float,)

        """
        # set/reset pop states, if needed:
        n_pop = len(vars_float)
        if not isinstance(self.algo.states, PopStates):
            self._reset_states(PopStates(self.algo.states, n_pop))
        elif self.algo.states.n_pop != n_pop:
            ostates = self.algo.states.states
            self._reset_states(PopStates(ostates, n_pop))

    def apply_individual(self, vars_int, vars_float):
        """
        Apply new variables to the problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        self._count += 1
        self.update_problem_individual(vars_int, vars_float)
        farm_results = self.algo.calc_farm(**self.calc_farm_args)

        if self.points is None:
            return farm_results
        else:
            point_results = self.algo.calc_points(farm_results, self.points)
            return farm_results, point_results

    def apply_population(self, vars_int, vars_float):
        """
        Apply new variables to the problem,
        for a whole population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        self._count += 1

        self.update_problem_population(vars_int, vars_float)
        farm_results = self.algo.calc_farm(**self.calc_farm_args)
        farm_results["n_pop"] = len(vars_float)
        farm_results["n_org_states"] = self._org_n_states

        if self.points is None:
            return farm_results
        else:
            n_pop = farm_results["n_pop"].values
            n_states, n_points = self.points.shape[:2]
            pop_points = np.zeros(
                (n_pop, n_states, n_points, 3), dtype=config.dtype_double
            )
            pop_points[:] = self.points[None, :, :, :]
            pop_points = pop_points.reshape(n_pop * n_states, n_points, 3)
            point_results = self.algo.calc_points(farm_results, pop_points)
            return farm_results, point_results

    def add_to_layout_figure(self, ax, **kwargs):
        """
        Add to a layout figure

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The figure axis

        """
        for c in self.cons.functions:
            ax = c.add_to_layout_figure(ax, **kwargs)
        for f in self.objs.functions:
            ax = f.add_to_layout_figure(ax, **kwargs)

        return ax



"""
Abstract classes and core functionality for wind farm optimization.
"""

from .farm_opt_problem import FarmOptProblem
from .farm_vars_problem import FarmVarsProblem
from .farm_objective import FarmObjective
from .farm_constraint import FarmConstraint
from .pop_states import PopStates



import numpy as np
import argparse
import matplotlib.pyplot as plt
from iwopy.interfaces.pymoo import Optimizer_pymoo

import foxes
from foxes_opt.problems.layout import RegularLayoutOptProblem
from foxes_opt.objectives import MaxFarmPower

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear_k002"],
        nargs="+",
    )
    parser.add_argument("-p", "--pwakes", help="The partial wakes model", default=None)
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.08)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-d",
        "--min_dist",
        help="Minimal turbine distance in m",
        type=float,
        default=500.0,
    )
    parser.add_argument(
        "-A", "--opt_algo", help="The pymoo algorithm name", default="GA"
    )
    parser.add_argument(
        "-P", "--n_pop", help="The population size", type=int, default=50
    )
    parser.add_argument(
        "-G", "--n_gen", help="The nmber of generations", type=int, default=150
    )
    parser.add_argument(
        "-nop", "--no_pop", help="Switch off vectorization", action="store_true"
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    boundary = foxes.utils.geom2d.Circle(
        [0.0, 0.0], 1000.0
    ) + foxes.utils.geom2d.ClosedPolygon(
        np.array([[0, 0], [0, 1600], [1000, 1800], [900, -200]], dtype=np.float64)
    )

    farm = foxes.WindFarm(boundary=boundary)
    farm.add_turbine(
        foxes.Turbine(
            xy=np.array([0.0, 0.0]), turbine_models=["layout_opt", "kTI_02", ttype.name]
        )
    )

    states = foxes.input.states.SingleStateStates(
        ws=args.ws, wd=args.wd, ti=args.ti, rho=args.rho
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
        verbosity=0,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
        verbosity=0,
    ):
        problem = RegularLayoutOptProblem("layout_opt", algo, min_spacing=args.min_dist)
        problem.add_objective(MaxFarmPower(problem))
        problem.initialize()

        solver = Optimizer_pymoo(
            problem,
            problem_pars=dict(
                vectorize=not args.no_pop,
            ),
            algo_pars=dict(
                type=args.opt_algo,
                pop_size=args.n_pop,
                seed=None,
            ),
            setup_pars=dict(),
            term_pars=("n_gen", args.n_gen),
        )
        solver.initialize()
        solver.print_info()

        ax = foxes.output.FarmLayoutOutput(farm).get_figure()
        plt.show()
        plt.close(ax.get_figure())

        results = solver.solve()
        solver.finalize(results)

        print()
        print(results)

        fig, axs = plt.subplots(1, 2, figsize=(12, 8))

        foxes.output.FarmLayoutOutput(farm).get_figure(fig=fig, ax=axs[0])

        o = foxes.output.FlowPlots2D(algo, results.problem_results)
        p_min = np.array([-1100.0, -1100.0])
        p_max = np.array([1100.0, 2000.0])
        fig = o.get_mean_fig_xy(
            "WS",
            resolution=20,
            fig=fig,
            ax=axs[1],
            xmin=p_min[0],
            xmax=p_max[0],
            ymin=p_min[1],
            ymax=p_max[1],
        )
        dpars = dict(alpha=0.6, zorder=10, p_min=p_min, p_max=p_max)
        farm.boundary.add_to_figure(
            axs[1], fill_mode="outside_white", pars_distance=dpars
        )

        plt.show()
        plt.close(fig)



"""
Purely geometric wind farm layout problems.
"""

from .geom_layout import GeomLayout
from .geom_reggrid import GeomRegGrid
from .geom_layout_gridded import GeomLayoutGridded
from .geom_reggrids import GeomRegGrids
from .objectives import OMaxN, OMinN, OFixN, MaxGridSpacing, MaxDensity, MeMiMaDist
from .constraints import Valid, Boundary, MinDist, CMinN, CMaxN, CFixN, CMinDensity



import numpy as np
import argparse
import matplotlib.pyplot as plt
from iwopy.interfaces.pymoo import Optimizer_pymoo

import foxes
import foxes_opt.problems.layout.geom_layouts as grg

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--min_dist",
        help="Minimal turbine distance in m",
        type=float,
        default=400.0,
    )
    parser.add_argument(
        "-m", "--n_maxr", help="Maximal turbines per row", type=int, default=None
    )
    parser.add_argument(
        "-g", "--n_grids", help="The number of grids", type=int, default=2
    )
    parser.add_argument(
        "-nop", "--no_pop", help="Switch off vectorization", action="store_true"
    )
    parser.add_argument(
        "-A", "--opt_algo", help="The pymoo algorithm name", default="MixedVariableGA"
    )
    parser.add_argument(
        "-P", "--n_pop", help="The population size", type=int, default=100
    )
    parser.add_argument(
        "-G", "--n_gen", help="The number of generations", type=int, default=150
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    args = parser.parse_args()

    boundary = (
        foxes.utils.geom2d.ClosedPolygon(
            np.array(
                [[0, 1000], [400, 1600], [2500, 1100], [2200, 300]], dtype=np.float64
            )
        )
        + foxes.utils.geom2d.Circle([2500.0, 0.0], 500.0)
        + foxes.utils.geom2d.ClosedPolygon(
            np.array(
                [[2200, 0], [2200, 2600], [2800, 2600], [2800, 0]], dtype=np.float64
            )
        )
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
        verbosity=0,
    ):
        problem = grg.GeomRegGrids(boundary, args.min_dist, args.n_grids, args.n_maxr)
        problem.add_objective(grg.OMaxN(problem))
        problem.initialize()

        fig = problem.get_fig().get_figure()
        plt.show()
        plt.close(fig)

        solver = Optimizer_pymoo(
            problem,
            problem_pars=dict(
                vectorize=not args.no_pop,
            ),
            algo_pars=dict(
                type=args.opt_algo,
                pop_size=args.n_pop,
                seed=None,
            ),
            setup_pars=dict(),
            term_pars=("n_gen", args.n_gen),
        )
        solver.initialize()
        solver.print_info()

        results = solver.solve()
        solver.finalize(results)

        print()
        print(results)

        xy, valid = results.problem_results
        fig = problem.get_fig(xy, valid).get_figure()
        plt.show()
        plt.close(fig)



import numpy as np

from foxes_opt.core.farm_constraint import FarmConstraint
import foxes.variables as FV
import foxes.constants as FC


class MinDistConstraint(FarmConstraint):
    """
    Turbines must keep at least a minimal
    spatial distance.

    Attributes
    ----------
    farm: foxes.WindFarm
        The wind farm
    sel_turbines: list
        The selected turbines
    min_dist: float
        The minimal distance
    min_dist_unit: str
        The minimal distance unit, either m or D

    :group: opt.constraints

    """

    def __init__(
        self,
        problem,
        min_dist,
        min_dist_unit="m",
        name="dist",
        sel_turbines=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying optimization problem
        min_dist: float
            The minimal distance
        min_dist_unit: str
            The minimal distance unit, either m or D
        name: str
            The name of the constraint
        sel_turbines: list of int, optional
            The selected turbines
        kwargs: dict, optional
            Additional parameters for `iwopy.Constraint`

        """
        self.min_dist = min_dist
        self.min_dist_unit = min_dist_unit

        selt = problem.sel_turbines if sel_turbines is None else sel_turbines
        vrs = []
        for ti in selt:
            vrs += [problem.tvar(FV.X, ti), problem.tvar(FV.Y, ti)]

        super().__init__(problem, name, sel_turbines, vnames_float=vrs, **kwargs)

    def initialize(self, verbosity=0):
        """
        Initialize the constaint.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        N = self.farm.n_turbines
        self._i2t = []  # i --> (ti, tj)
        self._t2i = np.full([N, N], -1)  # (ti, tj) --> i
        i = 0
        for ti in self.sel_turbines:
            for tj in range(N):
                if ti != tj and self._t2i[ti, tj] < 0:
                    self._i2t.append([ti, tj])
                    self._t2i[ti, tj] = i
                    self._t2i[tj, ti] = i
                    i += 1
        self._i2t = np.array(self._i2t)
        self._cnames = [f"{self.name}_{ti}_{tj}" for ti, tj in self._i2t]
        super().initialize(verbosity)

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return len(self._i2t)

    def vardeps_float(self):
        """
        Gets the dependencies of all components
        on the function float variables

        Returns
        -------
        deps: numpy.ndarray of bool
            The dependencies of components on function
            variables, shape: (n_components, n_vars_float)

        """
        turbs = list(self.problem.sel_turbines)
        deps = np.zeros((self.n_components(), len(turbs), 2), dtype=bool)
        for i, titj in enumerate(self._i2t):
            for t in titj:
                if t in turbs:
                    j = turbs.index(t)
                    deps[i, j] = True
        return deps.reshape(self.n_components(), 2 * len(turbs))

    def calc_individual(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_sel_components,)

        """
        xy = np.stack(
            [problem_results[FV.X].to_numpy(), problem_results[FV.Y].to_numpy()],
            axis=-1,
        )
        if not np.all(np.abs(np.min(xy, axis=0) - np.max(xy, axis=0)) < 1e-13):
            raise ValueError(f"Constraint '{self.name}': Require state independet XY")
        xy = xy[0]

        s = np.s_[:]
        if components is not None and len(components) < self.n_components():
            s = components

        a = np.take_along_axis(xy, self._i2t[s][:, 0, None], axis=0)
        b = np.take_along_axis(xy, self._i2t[s][:, 1, None], axis=0)
        d = np.linalg.norm(a - b, axis=-1)

        if self.min_dist_unit == "m":
            mind = self.min_dist

        elif self.min_dist_unit == "D":
            D = problem_results[FV.D].to_numpy()
            if not np.all(np.abs(np.min(D, axis=0) - np.max(D, axis=0)) < 1e-13):
                raise ValueError(
                    f"Constraint '{self.name}': Require state independet D"
                )
            D = D[0]

            Da = np.take_along_axis(D, self._i2t[s][:, 0], axis=0)
            Db = np.take_along_axis(D, self._i2t[s][:, 1], axis=0)
            mind = self.min_dist * np.maximum(Da, Db)

        return mind - d

    def calc_population(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        n_pop = problem_results["n_pop"].values
        n_states = problem_results["n_org_states"].values
        n_turbines = problem_results.sizes[FC.TURBINE]

        xy = np.stack(
            [problem_results[FV.X].to_numpy(), problem_results[FV.Y].to_numpy()],
            axis=-1,
        )
        xy = xy.reshape(n_pop, n_states, n_turbines, 2)
        if not np.all(np.abs(np.min(xy, axis=1) - np.max(xy, axis=1)) < 1e-13):
            raise ValueError(f"Constraint '{self.name}': Require state independet XY")
        xy = xy[:, 0]

        s = np.s_[:]
        if components is not None and len(components) < self.n_components():
            s = components

        a = np.take_along_axis(xy, self._i2t[s][None, :, 0, None], axis=1)
        b = np.take_along_axis(xy, self._i2t[s][None, :, 1, None], axis=1)
        d = np.linalg.norm(a - b, axis=-1)

        if self.min_dist_unit == "m":
            mind = self.min_dist

        elif self.min_dist_unit == "D":
            D = problem_results[FV.D].to_numpy().reshape(n_pop, n_states, n_turbines)
            if not np.all(np.abs(np.min(D, axis=1) - np.max(D, axis=1)) < 1e-13):
                raise ValueError(
                    f"Constraint '{self.name}': Require state independet D"
                )
            D = D[:, 0]

            Da = np.take_along_axis(D, self._i2t[s][None, :, 0], axis=1)
            Db = np.take_along_axis(D, self._i2t[s][None, :, 1], axis=1)
            mind = self.min_dist * np.maximum(Da, Db)

        return mind - d



import numpy as np

from foxes_opt.core.farm_constraint import FarmConstraint
import foxes.variables as FV


class AreaGeometryConstraint(FarmConstraint):
    """
    Constrains turbine positions to the inside
    of a given area geometry.

    Attributes
    ----------
    farm: foxes.WindFarm
        The wind farm
    sel_turbines: list
        The selected turbines
    geometry: foxes.utils.geom2d.AreaGeometry
        The area geometry
    disc_inside: bool
        Ensure full rotor disc inside boundary
    D: float
        Use this radius for rotor disc inside condition

    :group: opt.constraints

    """

    def __init__(
        self,
        problem,
        name,
        geometry,
        sel_turbines=None,
        disc_inside=False,
        D=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        problem : foxes_opt.FarmOptProblem
            The underlying optimization problem
        name : str
            The name of the constraint
        geometry : foxes.utils.geom2d.AreaGeometry
            The area geometry
        sel_turbines : list of int, optional
            The selected turbines
        disc_inside : bool
            Ensure full rotor disc inside boundary
        D : float, optional
            Use this radius for rotor disc inside condition
        kwargs : dict, optional
            Additional parameters for `iwopy.Constraint`

        """
        self.geometry = geometry
        self.disc_inside = disc_inside
        self.D = D

        selt = problem.sel_turbines if sel_turbines is None else sel_turbines
        vrs = []
        cns = []
        for ti in selt:
            vrs += [problem.tvar(FV.X, ti), problem.tvar(FV.Y, ti)]
            cns.append(f"{name}_{ti:04d}")

        super().__init__(
            problem, name, sel_turbines, vnames_float=vrs, cnames=cns, **kwargs
        )

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return self.n_sel_turbines

    def vardeps_float(self):
        """
        Gets the dependencies of all components
        on the function float variables

        Returns
        -------
        deps: numpy.ndarray of bool
            The dependencies of components on function
            variables, shape: (n_components, n_vars_float)

        """
        deps = np.zeros((self.n_components(), self.n_components(), 2), dtype=bool)
        np.fill_diagonal(deps[:, :, 0], True)
        np.fill_diagonal(deps[:, :, 1], True)
        return deps.reshape(self.n_components(), self.n_components() * 2)

    def calc_individual(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_sel_components,)

        """
        s = np.s_[:]
        if components is not None and len(components) < self.n_components():
            s = components
        xy = vars_float.reshape(self.n_components(), 2)[s]

        dists = self.geometry.points_distance(xy)
        dists[self.geometry.points_inside(xy)] *= -1

        if self.disc_inside:
            if self.D is None:
                dists += problem_results[FV.D].to_numpy()[0, self.sel_turbines][s] / 2
            else:
                dists += self.D / 2

        return dists

    def calc_population(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        n_pop = len(vars_float)
        n_cmpnts = self.n_components()
        s = np.s_[:]
        if components is not None and len(components) < self.n_components():
            n_cmpnts = len(components)
            s = components
        xy = vars_float[:, s].reshape(n_pop * n_cmpnts, 2)

        dists = self.geometry.points_distance(xy)
        dists[self.geometry.points_inside(xy)] *= -1
        dists = dists.reshape(n_pop, n_cmpnts)

        if self.disc_inside:
            if self.D is None:
                dists += (
                    problem_results[FV.D].to_numpy()[None, 0, self.sel_turbines][s] / 2
                )
            else:
                dists += self.D / 2

        return dists


class FarmBoundaryConstraint(AreaGeometryConstraint):
    """
    Constrains turbine positions to the inside of
    the wind farm boundary

    :group: opt.constraints

    """

    def __init__(self, problem, name="boundary", **kwargs):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying optimization problem
        name: str
            The name of the constraint
        kwargs: dict, optional
            Additional parameters for `AreaGeometryConstraint`

        """
        b = problem.farm.boundary
        assert b is not None, f"Constraint '{name}': Missing wind farm boundary."
        super().__init__(problem, name, geometry=b, **kwargs)



import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist
from iwopy import Problem

from foxes.config import config


class GeomRegGrid(Problem):
    """
    A regular grid within a boundary geometry.

    This optimization problem does not involve
    wind farms.

    Attributes
    ----------
    boundary: foxes.utils.geom2d.AreaGeometry
        The boundary geometry
    n_turbines: int
        The number of turbines in the layout
    min_dist: float
        The minimal distance between points
    max_dist: float
        The maximal distance between points
    D: float
        The diameter of circle fully within boundary

    :group: opt.problems.layout.geom_layouts

    """

    def __init__(
        self,
        boundary,
        n_turbines,
        min_dist,
        max_dist=None,
        D=None,
    ):
        """
        Constructor.

        Parameters
        ----------
        boundary: foxes.utils.geom2d.AreaGeometry
            The boundary geometry
        n_turbines: int
            The number of turbines in the layout
        min_dist: float
            The minimal distance between points
        max_dist: float, optional
            The maximal distance between points
        D: float, optional
            The diameter of circle fully within boundary

        """
        super().__init__(name="geom_reg_grid")

        self.boundary = boundary
        self.n_turbines = n_turbines
        self.min_dist = float(min_dist)
        self.max_dist = float(max_dist) if max_dist is not None else max_dist
        self.D = D

        self._SX = "sx"
        self._SY = "sy"
        self._DX = "dx"
        self._DY = "dy"
        self._ALPHA = "alpha"

    def initialize(self, verbosity=1):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(verbosity)

        pmin = self.boundary.p_min()
        pmax = self.boundary.p_max()
        self._pc = 0.5 * (pmin + pmax)
        self._diag = np.linalg.norm(pmax - pmin)
        self.max_dist = self._diag if self.max_dist is None else self.max_dist
        self._nrow = (
            int(np.maximum(self._diag / self.min_dist, np.sqrt(self.n_turbines) + 0.5))
            + 3
        )

        if verbosity > 0:
            print(f"Grid data:")
            print(f"  pmin        = {pmin}")
            print(f"  pmax        = {pmax}")
            print(f"  min dist    = {self.min_dist}")
            print(f"  max dist    = {self.max_dist}")
            print(f"  n row max   = {self._nrow}")
            print(f"  n max       = {self._nrow**2}")

        self.apply_individual(self.initial_values_int(), self.initial_values_float())

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        return list(np.array([self._SX, self._SY, self._DX, self._DY, self._ALPHA]))

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        vals = np.zeros(5, dtype=config.dtype_double)
        vals[2:4] = self.min_dist
        return vals

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        vals = np.zeros(5, dtype=config.dtype_double)
        vals[:2] = -0.5
        vals[2:4] = self.min_dist
        return vals

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        vals = np.zeros(5, dtype=config.dtype_double)
        vals[:2] = 0.5
        vals[2:4] = self.max_dist
        vals[4] = 90.0
        return vals

    def apply_individual(self, vars_int, vars_float):
        """
        Apply new variables to the problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        sx, sy, dx, dy, alpha = vars_float

        a = np.deg2rad(alpha)
        nax = np.stack([np.cos(a), np.sin(a)], axis=-1)
        nay = np.stack([-np.sin(a), np.cos(a)], axis=-1)

        pts = (
            self._pc[None, None, :]
            + (np.arange(self._nrow)[:, None, None] - (self._nrow - 1) / 2 + sx)
            * dx
            * nax[None, None, :]
            + (np.arange(self._nrow)[None, :, None] - (self._nrow - 1) / 2 + sy)
            * dy
            * nay[None, None, :]
        )
        pts = pts.reshape(self._nrow**2, 2)

        if self.D is None:
            valid = self.boundary.points_inside(pts)
        else:
            valid = self.boundary.points_inside(pts) & (
                self.boundary.points_distance(pts) >= self.D / 2
            )

        nvl = np.sum(valid)
        if nvl >= self.n_turbines:
            return pts[valid][: self.n_turbines], np.ones(self.n_turbines, dtype=bool)
        else:
            qts = np.append(pts[valid], pts[~valid][: (self.n_turbines - nvl)], axis=0)
            vld = np.zeros(self.n_turbines, dtype=bool)
            vld[:nvl] = True
            return qts, vld

    def apply_population(self, vars_int, vars_float):
        """
        Apply new variables to the problem,
        for a whole population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        n_pop = vars_float.shape[0]
        sx = vars_float[:, 0]
        sy = vars_float[:, 1]
        dx = vars_float[:, 2]
        dy = vars_float[:, 3]
        alpha = vars_float[:, 4]

        a = np.deg2rad(alpha)
        nax = np.stack([np.cos(a), np.sin(a)], axis=-1)
        nay = np.stack([-np.sin(a), np.cos(a)], axis=-1)

        pts = (
            self._pc[None, None, None, :]
            + (
                np.arange(self._nrow)[None, :, None, None]
                - (self._nrow - 1) / 2
                + sx[:, None, None, None]
            )
            * dx[:, None, None, None]
            * nax[:, None, None, :]
            + (
                np.arange(self._nrow)[None, None, :, None]
                - (self._nrow - 1) / 2
                + sy[:, None, None, None]
            )
            * dy[:, None, None, None]
            * nay[:, None, None, :]
        )
        pts = pts.reshape(n_pop * self._nrow**2, 2)

        if self.D is None:
            valid = self.boundary.points_inside(pts)
        else:
            valid = self.boundary.points_inside(pts) & (
                self.boundary.points_distance(pts) >= self.D / 2
            )
        valid = valid.reshape(n_pop, self._nrow**2)
        pts = pts.reshape(n_pop, self._nrow**2, 2)

        nvl = np.sum(valid, axis=1)
        qts = np.zeros((n_pop, self.n_turbines, 2), dtype=config.dtype_double)
        vld = np.zeros((n_pop, self.n_turbines), dtype=bool)
        for pi in range(n_pop):
            if nvl[pi] >= self.n_turbines:
                qts[pi] = pts[pi, valid[pi]][: self.n_turbines]
                vld[pi] = np.ones(self.n_turbines, dtype=bool)
            else:
                qts[pi] = np.append(
                    pts[pi, valid[pi]],
                    pts[pi, ~valid[pi]][: (self.n_turbines - nvl[pi])],
                    axis=0,
                )
                vld[pi, : nvl[pi]] = True

        return qts, vld

    def get_fig(
        self, xy=None, valid=None, ax=None, title=None, true_circle=True, **bargs
    ):
        """
        Return plotly figure axis.

        Parameters
        ----------
        xy: numpy.ndarary, optional
            The xy coordinate array, shape: (n_points, 2)
        valid: numpy.ndarray, optional
            Boolean array of validity, shape: (n_points,)
        ax: pyplot.Axis, optional
            The figure axis
        title: str, optional
            The figure title
        true_circle: bool
            Draw points as circles with diameter self.D
        bars: dict, optional
            The boundary plot arguments

        Returns
        -------
        ax: pyplot.Axis
            The figure axis

        """
        if ax is None:
            __, ax = plt.subplots()

        hbargs = {"fill_mode": "inside_lightgray"}
        hbargs.update(bargs)
        self.boundary.add_to_figure(ax, **hbargs)

        if xy is not None:
            if valid is not None:
                xy = xy[valid]
            if not true_circle or self.D is None:
                ax.scatter(xy[:, 0], xy[:, 1], color="orange")
            else:
                for x, y in xy:
                    ax.add_patch(
                        plt.Circle((x, y), self.D / 2, color="blue", fill=True)
                    )

        ax.set_aspect("equal", adjustable="box")
        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")

        if title is None:
            if xy is None:
                title = f"Optimization area"
            else:
                l = len(xy) if xy is not None else 0
                dists = cdist(xy, xy)
                np.fill_diagonal(dists, 1e20)
                title = f"N = {l}, min_dist = {np.min(dists):.1f} m"
        ax.set_title(title)

        return ax



import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist
from iwopy import Problem

from foxes.config import config


class GeomLayout(Problem):
    """
    A layout within a boundary geometry, purely
    defined by geometrical optimization (no wakes).

    This optimization problem does not involve
    wind farms.

    Attributes
    ----------
    boundary: foxes.utils.geom2d.AreaGeometry
        The boundary geometry
    n_turbines: int
        The number of turbines in the layout
    min_dist: float
        The minimal distance between points
    D: float
        The diameter of circle fully within boundary
    calc_valid: bool
        Evaluate validity

    :group: opt.problems.layout.geom_layouts

    """

    def __init__(
        self,
        boundary,
        n_turbines,
        min_dist=None,
        D=None,
        calc_valid=None,
    ):
        """
        Constructor.

        Parameters
        ----------
        boundary: foxes.utils.geom2d.AreaGeometry
            The boundary geometry
        n_turbines: int
            The number of turbines in the layout
        min_dist: float, optional
            The minimal distance between points
        D: float, optional
            The diameter of circle fully within boundary
        calc_valid: bool, optional
            Evaluate validity

        """
        super().__init__(name="geom_reg_grids")

        self.boundary = boundary
        self.n_turbines = n_turbines
        self.D = D
        self.min_dist = min_dist
        self.calc_valid = calc_valid
        if calc_valid is None:
            self.calc_valid = min_dist is not None or D is not None

        self._X = [f"x{i}" for i in range(self.n_turbines)]
        self._Y = [f"y{i}" for i in range(self.n_turbines)]

    def initialize(self, verbosity=1):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(verbosity)
        self.apply_individual(self.initial_values_int(), self.initial_values_float())

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        return list(np.array([self._X, self._Y]).T.flat)

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        pmin = self.boundary.p_min()
        pmax = self.boundary.p_max()
        pc = 0.5 * (pmin + pmax)
        delta = 0.8 * (pmax - pmin)

        vals = np.zeros((self.n_turbines, 2), dtype=config.dtype_double)
        vals[:] = pc[None, :] - 0.5 * delta[None, :]
        vals[:] += (
            np.arange(self.n_turbines)[:, None] * delta[None, :] / (self.n_turbines - 1)
        )

        return vals.reshape(self.n_turbines * 2)

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        vals = np.zeros((self.n_turbines, 2), dtype=config.dtype_double)
        vals[:] = self.boundary.p_min()[None, :]
        return vals.reshape(self.n_turbines * 2)

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        vals = np.zeros((self.n_turbines, 2), dtype=config.dtype_double)
        vals[:] = self.boundary.p_max()[None, :]
        return vals.reshape(self.n_turbines * 2)

    def apply_individual(self, vars_int, vars_float):
        """
        Apply new variables to the problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        xy = vars_float.reshape(self.n_turbines, 2)

        valid = None
        if self.calc_valid:
            if self.D is None:
                valid = self.boundary.points_inside(xy)
            else:
                valid = self.boundary.points_inside(xy) & (
                    self.boundary.points_distance(xy) >= self.D / 2
                )

            if self.min_dist is not None:
                dists = cdist(xy, xy)
                np.fill_diagonal(dists, 1e20)
                dists = np.min(dists, axis=1)
                valid[dists < self.min_dist] = False

        return xy, valid

    def apply_population(self, vars_int, vars_float):
        """
        Apply new variables to the problem,
        for a whole population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        n_pop = vars_float.shape[0]
        xy = vars_float.reshape(n_pop, self.n_turbines, 2)

        valid = None
        if self.calc_valid:
            qts = xy.reshape(n_pop * self.n_turbines, 2)
            if self.D is None:
                valid = self.boundary.points_inside(qts)
            else:
                valid = self.boundary.points_inside(qts) & (
                    self.boundary.points_distance(qts) >= self.D / 2
                )
            valid = valid.reshape(n_pop, self.n_turbines)

            if self.min_dist is not None:
                for pi in range(n_pop):
                    dists = cdist(xy[pi], xy[pi])
                    np.fill_diagonal(dists, 1e20)
                    dists = np.min(dists, axis=1)
                    valid[pi, dists < self.min_dist] = False

        return xy, valid

    def get_fig(
        self, xy=None, valid=None, ax=None, title=None, true_circle=True, **bargs
    ):
        """
        Return plotly figure axis.

        Parameters
        ----------
        xy: numpy.ndarary, optional
            The xy coordinate array, shape: (n_points, 2)
        valid: numpy.ndarray, optional
            Boolean array of validity, shape: (n_points,)
        ax: pyplot.Axis, optional
            The figure axis
        title: str, optional
            The figure title
        true_circle: bool
            Draw points as circles with diameter self.D
        bars: dict, optional
            The boundary plot arguments

        Returns
        -------
        ax: pyplot.Axis
            The figure axis

        """
        if ax is None:
            __, ax = plt.subplots()

        hbargs = {"fill_mode": "inside_lightgray"}
        hbargs.update(bargs)
        self.boundary.add_to_figure(ax, **hbargs)

        if xy is not None:
            if valid is not None:
                xy = xy[valid]
            if not true_circle or self.D is None:
                ax.scatter(xy[:, 0], xy[:, 1], color="orange")
            else:
                for x, y in xy:
                    ax.add_patch(
                        plt.Circle((x, y), self.D / 2, color="blue", fill=True)
                    )

        ax.set_aspect("equal", adjustable="box")
        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")

        if title is None:
            if xy is None:
                title = f"Optimization area"
            else:
                l = len(xy) if xy is not None else 0
                dists = cdist(xy, xy)
                np.fill_diagonal(dists, 1e20)
                title = f"N = {l}, min_dist = {np.min(dists):.1f} m"
        ax.set_title(title)

        return ax



import numpy as np
import argparse
import matplotlib.pyplot as plt
from iwopy import LocalFD
from iwopy.optimizers import GG

import foxes
from foxes_opt.problems import OptFarmVars
from foxes_opt.objectives import MaxFarmPower
import foxes.variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-nt", "--n_t", help="The number of turbines", type=int, default=9
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["CrespoHernandez_quadratic", "Bastankhah2016_linear"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=["kTI_02"], nargs="+"
    )
    parser.add_argument("-p", "--pwakes", help="The partial wakes model", default=None)
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.03)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-d",
        "--min_dist",
        help="Minimal turbine distance in unit D",
        type=float,
        default=None,
    )
    parser.add_argument(
        "-nop", "--no_pop", help="Switch off vectorization", action="store_true"
    )
    parser.add_argument(
        "-O",
        "--fd_order",
        help="Finite difference derivative order",
        type=int,
        default=1,
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    farm = foxes.WindFarm()
    N = int(np.sqrt(args.n_t) + 0.5)
    foxes.input.farm_layout.add_grid(
        farm,
        xy_base=np.array([500.0, 500.0]),
        step_vectors=np.array([[1300.0, 0], [200, 600.0]]),
        steps=(N, N),
        turbine_models=args.tmodels + ["opt_yawm", "yawm2yaw", ttype.name],
    )
    states = foxes.input.states.SingleStateStates(
        ws=args.ws, wd=args.wd, ti=args.ti, rho=args.rho
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="yawed",
        partial_wakes=args.pwakes,
        mbook=mbook,
        verbosity=0,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
        verbosity=0,
    ):
        problem = OptFarmVars("opt_yawm", algo)
        problem.add_var(FV.YAWM, float, 0.0, -40.0, 40.0, level="turbine")
        problem.add_objective(MaxFarmPower(problem))
        problem.initialize()
        gproblem = LocalFD(problem, deltas=0.1, fd_order=args.fd_order)
        gproblem.initialize()

        solver = GG(
            gproblem,
            step_max=100.0,
            step_min=0.1,
            f_tol=1e-4,
            vectorized=not args.no_pop,
        )
        solver.initialize()
        solver.print_info()

        ax = foxes.output.FarmLayoutOutput(farm).get_figure()
        plt.show()
        plt.close(ax.get_figure())

        results = solver.solve()
        solver.finalize(results)

        print()
        print(results)

        fr = results.problem_results.to_dataframe()
        print(fr[[FV.X, FV.Y, FV.AMB_WD, FV.REWS, FV.TI, FV.P, FV.YAWM]])

        o = foxes.output.FlowPlots2D(algo, results.problem_results)
        fig = o.get_mean_fig_xy("WS", resolution=10)
        plt.show()



import numpy as np
from iwopy.interfaces.pymoo import Optimizer_pymoo

import foxes
from foxes_opt.problems.layout import FarmLayoutOptProblem
from foxes_opt.constraints import FarmBoundaryConstraint, MinDistConstraint
from foxes_opt.objectives import MaxFarmPower


def test():

    boundary = foxes.utils.geom2d.Circle([0.0, 0.0], 1000.0)

    farm = foxes.WindFarm(boundary=boundary)
    foxes.input.farm_layout.add_row(
        farm=farm,
        xy_base=np.zeros(2),
        xy_step=np.array([50.0, 0.0]),
        n_turbines=8,
        turbine_models=["NREL5MW"],
    )
    states = foxes.input.states.SingleStateStates(ws=9, wd=270, ti=0.08, rho=1.225)

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model="centre",
        wake_models=["Bastankhah2014_linear_lim_k004"],
        verbosity=0,
    )

    problem = FarmLayoutOptProblem("layout_opt", algo)
    problem.add_objective(MaxFarmPower(problem))
    problem.add_constraint(FarmBoundaryConstraint(problem, disc_inside=True))
    problem.add_constraint(MinDistConstraint(problem, min_dist=3, min_dist_unit="D"))
    problem.initialize()

    solver = Optimizer_pymoo(
        problem,
        problem_pars=dict(
            vectorize=True,
        ),
        algo_pars=dict(
            type="GA",
            pop_size=100,
            seed=41,
        ),
        setup_pars=dict(),
        term_pars=dict(
            type="default",
            n_max_gen=100,
            ftol=1e-6,
            xtol=1e-6,
        ),
    )
    solver.initialize()
    solver.print_info()

    """
    ax = foxes.output.FarmLayoutOutput(farm).get_figure()
    plt.show()
    plt.close(ax.get_figure())
    """

    results = solver.solve()
    solver.finalize(results)

    print()
    print(results)

    assert np.abs(results.objs - 20148.46492534) < 1e-4

    """
    fig, axs = plt.subplots(1, 2, figsize=(12, 8))
    foxes.output.FarmLayoutOutput(farm).get_figure(fig=fig, ax=axs[0])

    o = foxes.output.FlowPlots2D(algo, results.problem_results)
    p_min = np.array([-1100.0, -1100.0])
    p_max = np.array([1100.0, 1100.0])
    fig = o.get_mean_fig_xy(
        "WS",
        resolution=20,
        fig=fig,
        ax=axs[1],
        xmin=p_min[0],
        xmax=p_max[0],
        ymin=p_min[1],
        ymax=p_max[1],
    )
    
    dpars = dict(alpha=0.6, zorder=10, p_min=p_min, p_max=p_max)
    farm.boundary.add_to_figure(
        axs[1], fill_mode="outside_white", pars_distance=dpars
    )
    plt.show()
    plt.close(fig)
    """


if __name__ == "__main__":

    test()



import numpy as np
from scipy.spatial.distance import cdist
from iwopy import Constraint

from foxes.config import config


class Valid(Constraint):
    """
    Validity constraint for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.constraints

    """

    def __init__(self, problem, name="valid", **kwargs):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        name: str
            The constraint name
        kwargs: dict, optional
            Additioal parameters for the base class

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
            **kwargs,
        )

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        __, valid = problem_results
        return np.sum(~valid)

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        __, valid = problem_results
        return np.sum(~valid, axis=1)[:, None]


class Boundary(Constraint):
    """
    Boundary constraint for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.constraints

    """

    def __init__(self, problem, n_turbines=None, D=None, name="boundary", **kwargs):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        n_turbines: int, optional
            The number of turbines
        D: float, optional
            The rotor diameter
        name: str
            The constraint name
        kwargs: dict, optional
            Additioal parameters for the base class

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
            **kwargs,
        )
        self.n_turbines = problem.n_turbines if n_turbines is None else n_turbines
        self.D = problem.D if D is None else D

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return self.n_turbines

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        xy, __ = problem_results

        dists = self.problem.boundary.points_distance(xy)
        dists[self.problem.boundary.points_inside(xy)] *= -1

        if self.D is not None:
            dists += self.D / 2

        return dists

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        xy, __ = problem_results
        n_pop, n_xy = xy.shape[:2]

        xy = xy.reshape(n_pop * n_xy, 2)
        dists = self.problem.boundary.points_distance(xy)
        dists[self.problem.boundary.points_inside(xy)] *= -1
        dists = dists.reshape(n_pop, n_xy)

        if self.D is not None:
            dists += self.D / 2

        return dists


class MinDist(Constraint):
    """
    Minimal distance constraint for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.constraints

    """

    def __init__(
        self, problem, min_dist=None, n_turbines=None, name="min_dist", **kwargs
    ):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        min_dist: float, optional
            The minimal distance between turbines
        n_turbines: int, optional
            The number of turbines
        name: str
            The constraint name
        kwargs: dict, optional
            Additioal parameters for the base class

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
            **kwargs,
        )
        self.min_dist = problem.min_dist if min_dist is None else min_dist
        self.n_turbines = problem.n_turbines if n_turbines is None else n_turbines

    def initialize(self, verbosity=0):
        """
        Initialize the constaint.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        N = self.n_turbines
        self._i2t = []  # i --> (ti, tj)
        self._t2i = np.full([N, N], -1)  # (ti, tj) --> i
        i = 0
        for ti in range(N):
            for tj in range(N):
                if ti != tj and self._t2i[ti, tj] < 0:
                    self._i2t.append([ti, tj])
                    self._t2i[ti, tj] = i
                    self._t2i[tj, ti] = i
                    i += 1
        self._i2t = np.array(self._i2t)
        self._cnames = [f"{self.name}_{ti}_{tj}" for ti, tj in self._i2t]
        super().initialize(verbosity)

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return len(self._i2t)

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        xy, __ = problem_results

        a = np.take_along_axis(xy, self._i2t[:, 0, None], axis=0)
        b = np.take_along_axis(xy, self._i2t[:, 1, None], axis=0)
        d = np.linalg.norm(a - b, axis=-1)

        return self.min_dist - d

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        xy, __ = problem_results

        a = np.take_along_axis(xy, self._i2t[None, :, 0, None], axis=1)
        b = np.take_along_axis(xy, self._i2t[None, :, 1, None], axis=1)
        d = np.linalg.norm(a - b, axis=-1)

        return self.min_dist - d


class CMinN(Constraint):
    """
    Minimal number of turbines constraint for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.constraints

    """

    def __init__(self, problem, N, name="cminN", **kwargs):
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
            **kwargs,
        )
        """
        Constructor.
        
        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying geometrical layout 
            optimization problem
        N: int
            The minimal number of turbines
        name: str
            The constraint name
        kwargs: dict, optional
            Additioal parameters for the base class

        """
        self.N = N

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        __, valid = problem_results
        return self.N - np.sum(valid)

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        __, valid = problem_results
        return self.N - np.sum(valid, axis=1)[:, None]


class CMaxN(Constraint):
    """
    Maximal number of turbines constraint for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.constraints

    """

    def __init__(self, problem, N, name="cmaxN", **kwargs):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        N: int
            The maximal number of turbines
        name: str
            The constraint name
        kwargs: dict, optional
            Additioal parameters for the base class

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
            **kwargs,
        )
        self.N = N

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        __, valid = problem_results
        return np.sum(valid) - self.N

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        __, valid = problem_results
        return np.sum(valid, axis=1)[:, None] - self.N


class CFixN(Constraint):
    """
    Fixed number of turbines constraint for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.constraints

    """

    def __init__(self, problem, N, name="cfixN", **kwargs):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        N: int
            The number of turbines
        name: str
            The constraint name
        kwargs: dict, optional
            Additioal parameters for the base class

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
            tol=0.1,
            **kwargs,
        )
        self.N = N

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 2

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        __, valid = problem_results
        vld = np.sum(valid)
        return np.array([self.N - vld, vld - self.N])

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        __, valid = problem_results
        vld = np.sum(valid, axis=1)
        return np.stack([self.N - vld, vld - self.N], axis=-1)


class CMinDensity(Constraint):
    """
    Minimal turbine density constraint for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.constraints

    """

    def __init__(self, problem, min_value, dfactor=1, name="min_density"):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes_opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        min_value: float
            The minimal turbine density
        dfactor: float
            Delta factor for grid spacing
        name: str
            The constraint name
        kwargs: dict, optional
            Additioal parameters for the base class

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
        )
        self.min_value = min_value
        self.dfactor = dfactor

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def initialize(self, verbosity):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(verbosity)

        # define regular grid of probe points:
        geom = self.problem.boundary
        pmin = geom.p_min()
        pmax = geom.p_max()
        detlta = self.problem.min_dist / self.dfactor
        self._probes = np.stack(
            np.meshgrid(
                np.arange(pmin[0] - detlta, pmax[0] + 2 * detlta, detlta),
                np.arange(pmin[1] - detlta, pmax[1] + 2 * detlta, detlta),
                indexing="ij",
            ),
            axis=-1,
        )
        nx, ny = self._probes.shape[:2]
        n = nx * ny
        self._probes = self._probes.reshape(n, 2)

        # reduce to points within geometry:
        valid = geom.points_inside(self._probes)
        self._probes = self._probes[valid]

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        xy, valid = problem_results
        xy = xy[valid]
        dists = cdist(self._probes, xy)
        return np.nanmax(np.nanmin(dists, axis=1)) - self.min_value

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        n_pop = vars_float.shape[0]
        xy, valid = problem_results
        out = np.full(n_pop, 1e20, dtype=config.dtype_double)
        for pi in range(n_pop):
            if np.any(valid[pi]):
                hxy = xy[pi][valid[pi]]
                dists = cdist(self._probes, hxy)
                out[pi] = np.nanmax(np.nanmin(dists, axis=1)) - self.min_value
        return out[:, None]



"""
Wind farm optimization.
"""

from .core import FarmOptProblem, FarmObjective, FarmConstraint  # noqa: F401

from . import problems  # noqa: F401
from . import constraints  # noqa: F401
from . import objectives  # noqa: F401

from importlib.metadata import version

__version__ = version(__package__ or __name__)



import numpy as np
from copy import deepcopy

from foxes_opt.core import FarmOptProblem, FarmVarsProblem
from foxes.models.turbine_models import Calculator
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC

from .geom_layouts.geom_reggrids import GeomRegGrids


class RegGridsLayoutOptProblem(FarmVarsProblem):
    """
    Places turbines on several regular grids and optimizes
    their parameters.

    Note that this problem has both int and float variables
    (mixed problem).

    Attributes
    ----------
    min_spacing: float
        The minimal turbine spacing
    n_grids: int
        The number of grids
    max_n_row: int
        The maximal number of turbines per
        grid and row

    :group: opt.problems.layout

    """

    def __init__(
        self,
        name,
        algo,
        min_dist,
        n_grids=1,
        n_row_max=None,
        max_dist=None,
        runner=None,
        **kwargs,
    ):
        """
        Constraints.

        Parameters
        ----------
        name: str
            The problem's name
        algo: foxes.core.Algorithm
            The algorithm
        min_dist: float
            The minimal distance between points
        n_grids: int
            The number of grids
        n_row_max: int, optional
            The maximal number of points in a row
        max_dist: float, optional
            The maximal distance between points
        runner: foxes.core.Runner, optional
            The runner for running the algorithm
        kwargs: dict, optional
            Additional parameters for `FarmVarsProblem`

        """
        super().__init__(name, algo, runner, **kwargs)

        b = algo.farm.boundary
        assert b is not None, f"Problem '{self.name}': Missing wind farm boundary."

        self._geomp = GeomRegGrids(
            b,
            min_dist=min_dist,
            n_grids=n_grids,
            n_row_max=n_row_max,
            max_dist=max_dist,
        )

    def initialize(self, verbosity=1, **kwargs):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for super class init

        """
        self._geomp.objs = self.objs
        self._geomp.cons = self.cons
        self._geomp.initialize(verbosity)

        self._mname = self.name + "_calc"
        for t in self.algo.farm.turbines:
            if self._mname not in t.models:
                t.models.append(self._mname)
        self._turbine = deepcopy(self.farm.turbines[-1])

        self.algo.mbook.turbine_models[self._mname] = Calculator(
            in_vars=[FC.VALID, FV.P, FV.CT],
            out_vars=[FC.VALID, FV.P, FV.CT],
            func=lambda valid, P, ct, st_sel: (valid, P * valid, ct * valid),
            pre_rotor=False,
        )

        super().initialize(
            pre_rotor_vars=[FV.X, FV.Y, FC.VALID],
            post_rotor_vars=[],
            verbosity=verbosity,
            **kwargs,
        )

    def var_names_int(self):
        """
        The names of int variables.

        Returns
        -------
        names: list of str
            The names of the int variables

        """
        return self._geomp.var_names_int()

    def initial_values_int(self):
        """
        The initial values of the int variables.

        Returns
        -------
        values: numpy.ndarray
            Initial int values, shape: (n_vars_int,)

        """
        return self._geomp.initial_values_int()

    def min_values_int(self):
        """
        The minimal values of the integer variables.

        Use -self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal int values, shape: (n_vars_int,)

        """
        return self._geomp.min_values_int()

    def max_values_int(self):
        """
        The maximal values of the integer variables.

        Use self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal int values, shape: (n_vars_int,)

        """
        return self._geomp.max_values_int()

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        return self._geomp.var_names_float()

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        return self._geomp.initial_values_float()

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        return self._geomp.min_values_float()

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        return self._geomp.max_values_float()

    def update_problem_individual(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        """
        n0 = self.farm.n_turbines
        nxny = vars_int.reshape(self._geomp.n_grids, 2)
        n = np.sum(np.prod(nxny, axis=1))
        if n0 > n:
            self.farm.turbines = self.farm.turbines[:n]
        elif n0 < n:
            for i in range(n0, n):
                self.farm.turbines.append(deepcopy(self._turbine))
                self.farm.turbines[-1].index = n0 + i
                self.farm.turbines[-1].name = f"T{n0 + i}"
        if n != n0:
            self.algo.update_n_turbines()

        super().update_problem_individual(vars_int, vars_float)

    def update_problem_population(self, vars_int, vars_float):
        """
        Update the algo and other data using
        the latest optimization variables.

        This function is called before running the farm
        calculation.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float,)

        """
        n0 = self.farm.n_turbines
        n_pop = vars_int.shape[0]
        nxny = vars_int.reshape(n_pop, self._geomp.n_grids, 2)
        n = np.max(np.sum(np.prod(nxny, axis=2), axis=1))
        if n0 > n:
            self.farm.turbines = self.farm.turbines[:n]
        elif n0 < n:
            for i in range(n0, n):
                self.farm.turbines.append(deepcopy(self._turbine))
                self.farm.turbines[-1].index = n0 + i
                self.farm.turbines[-1].name = f"T{n0 + i}"
        if n != n0:
            self.algo.update_n_turbines()

        super().update_problem_population(vars_int, vars_float)

    def opt2farm_vars_individual(self, vars_int, vars_float):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_vars_int,)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_vars_float,)

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_states, n_sel_turbines)

        """
        pts, vld = self._geomp.apply_individual(vars_int, vars_float)

        n_pts = pts.shape[0]
        n_states = self.algo.n_states
        n_turbines = self.farm.n_turbines

        pmi = np.min(self._geomp._pmin)
        points = np.full((n_states, n_turbines, 2), pmi, dtype=config.dtype_double)
        if n_pts <= n_turbines:
            points[:, :n_pts] = pts[None, :, :]
        else:
            points[:] = pts[None, :n_turbines, :]

        valid = np.zeros((n_states, n_turbines), dtype=config.dtype_double)
        if n_pts <= n_turbines:
            valid[:, :n_pts] = vld[None, :]
        else:
            valid[:] = vld[None, :n_turbines]

        farm_vars = {FV.X: points[:, :, 0], FV.Y: points[:, :, 1], FC.VALID: valid}

        return farm_vars

    def opt2farm_vars_population(self, vars_int, vars_float, n_states):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_pop, n_vars_int)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_pop, n_vars_float)
        n_states: int
            The number of original (non-pop) states

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_pop, n_states, n_sel_turbines)

        """
        pts, vld = self._geomp.apply_population(vars_int, vars_float)

        n_pop, n_pts = vld.shape
        n_states = self._org_n_states
        n_turbines = self.farm.n_turbines

        pmi = np.min(self._geomp._pmin)
        points = np.full(
            (n_pop, n_states, n_turbines, 2), pmi, dtype=config.dtype_double
        )
        if n_pts <= n_turbines:
            points[:, :, :n_pts] = pts[:, None, :, :]
        else:
            points[:] = pts[:, None, :n_turbines, :]

        valid = np.zeros((n_pop, n_states, n_turbines), dtype=config.dtype_double)
        if n_pts <= n_turbines:
            valid[:, :, :n_pts] = vld[:, None, :]
        else:
            valid[:] = vld[:, None, :n_turbines]

        farm_vars = {
            FV.X: points[:, :, :, 0],
            FV.Y: points[:, :, :, 1],
            FC.VALID: valid,
        }

        return farm_vars

    def finalize_individual(self, vars_int, vars_float, verbosity=1):
        """
        Finalization, given the champion data.

        Parameters
        ----------
        vars_int: np.array
            The optimal integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The optimal float variable values, shape: (n_vars_float,)
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem
        objs: np.array
            The objective function values, shape: (n_objectives,)
        cons: np.array
            The constraints values, shape: (n_constraints,)

        """
        pts, vld = self._geomp.apply_individual(vars_int, vars_float)
        xy = pts[vld]
        n_xy = xy.shape[0]

        self.farm.turbines = self.farm.turbines[:n_xy]
        for ti, t in enumerate(self.farm.turbines):
            t.xy = xy[ti]
            t.index = ti
            t.name = f"T{ti}"
            t.models = [
                mname for mname in t.models if mname not in [self.name, self._mname]
            ]
        self.algo.update_n_turbines()

        return FarmOptProblem.finalize_individual(
            self, vars_int, vars_float, verbosity=1
        )
